-- Addon Declaration
local addon = LibStub("AceAddon-3.0"):NewAddon("HoTorDoT","AceEvent-3.0")

local candy = LibStub("LibCandyBar-3.0")
local media = LibStub("LibSharedMedia-3.0")
local mType = media.MediaType and media.MediaType.STATUSBAR or "statusbar"

local L = LibStub("AceLocale-3.0"):GetLocale("HoTorDoT")

local normalAnchor = nil

local db
local defaults = {
	profile = {
		scale = 1.0,
		texture = "Minimalist",
		growup = false,
		time = true,
		align = "LEFT",
		icon = true,
		minDuration = 600,
		HoTorDoTTargetAnchor_width = 200,
		colors = {
			barBackground = { 0.5, 0.5, 0.5, 0.3 },
			barText = { 1, 1, 1 },
			barColor = { 0.25, 0.33, 0.68, 1 },
		}
	}
}
local namedBars = {}

local function barSorter(a, b)
	return a.remaining < b.remaining and true or false
end
local tmp = {}
local function rearrangeBars(anchor)
	wipe(tmp)
	for bar in pairs(anchor.bars) do
		table.insert(tmp, bar)
	end
	table.sort(tmp, barSorter)
	local lastDownBar, lastUpBar = nil, nil
	local up = db.growup
	for i, bar in ipairs(tmp) do
		bar:ClearAllPoints()
		if up then
			bar:SetPoint("BOTTOMLEFT", lastUpBar or anchor, "TOPLEFT")
			bar:SetPoint("BOTTOMRIGHT", lastUpBar or anchor, "TOPRIGHT")
			lastUpBar = bar
		else
			bar:SetPoint("TOPLEFT", lastDownBar or anchor, "BOTTOMLEFT")
			bar:SetPoint("TOPRIGHT", lastDownBar or anchor, "BOTTOMRIGHT")
			lastDownBar = bar
		end
	end
end

local function barStopped(event, bar)
	local a = bar:Get("hotordot:anchor")
	if a and a.bars and a.bars[bar] then
		a.bars[bar] = nil
		rearrangeBars(a)
	end
	a = bar:Get("hotordot:identifier")
	if a then namedBars[a] = nil end
end

--------------------------------------------------------------------------------
-- Anchors
--
local defaultPositions = {
	HoTorDoTTargetAnchor = {"CENTER", "UIParent", "CENTER", 0, -150},
}

local function onDragHandleMouseDown(self) self:GetParent():StartSizing("BOTTOMRIGHT") end
local function onDragHandleMouseUp(self, button) self:GetParent():StopMovingOrSizing() end
local function onResize(self, width)
	db[self.w] = width
	rearrangeBars(self)
end
local function onDragStart(self) self:StartMoving() end
local function onDragStop(self)
	self:StopMovingOrSizing()
	local s = self:GetEffectiveScale()
	db[self.x] = self:GetLeft() * s
	db[self.y] = self:GetTop() * s
end

local function onControlEnter(self)
	GameTooltip:ClearLines()
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT")
	GameTooltip:AddLine(self.tooltipHeader)
	GameTooltip:AddLine(self.tooltipText, 1, 1, 1, 1)
	GameTooltip:Show()
end
local function onControlLeave() GameTooltip:Hide() end

--local function menu() dew:FeedAceOptionsTable(addon.consoleOptions) end
local function displayOnMouseDown(self, button)
	if button == "RightButton" then
		print("this should open the menu")
		--dew:Open(self, "children", menu)
	end
end

local function createAnchor(frameName, title)
	local display = CreateFrame("Frame", frameName, UIParent)
	local wKey, xKey, yKey = frameName .. "_width", frameName .. "_x", frameName .. "_y"
	display.w, display.x, display.y = wKey, xKey, yKey
	display:EnableMouse(true)
	display:SetClampedToScreen(true)
	display:SetMovable(true)
	display:SetResizable(true)
	display:RegisterForDrag("LeftButton")
	display:SetWidth(db[wKey] or 200)
	display:SetHeight(20)
	display:SetMinResize(80, 20)
	display:SetMaxResize(1920, 20)
	display:ClearAllPoints()
	if db[xKey] and db[yKey] then
		local s = display:GetEffectiveScale()
		display:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", db[xKey] / s, db[yKey] / s)
	else
		display:SetPoint(unpack(defaultPositions[frameName]))
	end
	local bg = display:CreateTexture(nil, "PARENT")
	bg:SetAllPoints(display)
	bg:SetBlendMode("BLEND")
	bg:SetTexture(0, 0, 0, 0.3)
	local header = display:CreateFontString(nil, "OVERLAY")
	header:SetFontObject(GameFontNormal)
	header:SetText(title)
	header:SetAllPoints(display)
	header:SetJustifyH("CENTER")
	header:SetJustifyV("MIDDLE")
	local drag = CreateFrame("Frame", nil, display)
	drag:SetFrameLevel(display:GetFrameLevel() + 10)
	drag:SetWidth(16)
	drag:SetHeight(16)
	drag:SetPoint("BOTTOMRIGHT", display, -1, 1)
	drag:EnableMouse(true)
	drag:SetScript("OnMouseDown", onDragHandleMouseDown)
	drag:SetScript("OnMouseUp", onDragHandleMouseUp)
	drag:SetAlpha(0.5)
	local tex = drag:CreateTexture(nil, "BACKGROUND")
	tex:SetTexture("Interface\\AddOns\\HoTorDoT\\textures\\draghandle")
	tex:SetWidth(16)
	tex:SetHeight(16)
	tex:SetBlendMode("ADD")
	tex:SetPoint("CENTER", drag)

	local close = CreateFrame("Button", nil, display)
	close:SetPoint("BOTTOMLEFT", display, "BOTTOMLEFT", 3, 3)
	close:SetHeight(14)
	close:SetWidth(14)
	close.tooltipHeader = L["Hide"]
	close.tooltipText = L["Hides the anchors."]
	close:SetScript("OnEnter", onControlEnter)
	close:SetScript("OnLeave", onControlLeave)
	close:SetScript("OnClick", function() addon:ShowAnchors() end)
	close:SetNormalTexture("Interface\\AddOns\\HoTorDoT\\textures\\close")
	display:SetScript("OnSizeChanged", onResize)
	display:SetScript("OnDragStart", onDragStart)
	display:SetScript("OnDragStop", onDragStop)
	display:SetScript("OnMouseDown", displayOnMouseDown)
	display.bars = {}
	display:Hide()
	return display
end

function addon:ShowAnchors()
	if normalAnchor:IsShown() then
		normalAnchor:Hide()
		-- emphasizeAnchor:Hide()
	else
		normalAnchor:Show()
		-- emphasizeAnchor:Show()
	end
end

function addon:ResetAnchors()
	normalAnchor:ClearAllPoints()
	normalAnchor:SetPoint(unpack(defaultPositions[normalAnchor:GetName()]))
	db[normalAnchor.x] = nil
	db[normalAnchor.y] = nil
	db[normalAnchor.w] = nil
	normalAnchor:SetWidth(self.defaultDB[normalAnchor.w])
end


function addon:OnInitialize()
	candy.RegisterCallback(self, "LibCandyBar_Stop", barStopped)
	
	self.db = LibStub("AceDB-3.0"):New("HoTorDoTDB", defaults, "Default")
	db = self.db.profile
end

function addon:OnEnable()
	normalAnchor = createAnchor("HoTorDoTTargetAnchor", L["Target Bars"])
	if not media:Fetch(mType, db.texture, true) then db.texture = "BantoBar" end

	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
end

--------------------------------------------------------------------------------
-- Colors
--

local function colorNormal() return unpack(db.colors.barColor) end
local function colorText() return unpack(db.colors.barText) end
local function colorBackground() return unpack(db.colors.barBackground) end

--------------------------------------------------------------------------------
-- Event Handlers
--

local function start(text, rank, icon, time, duration, expirationTime, debuffType)
	local bar = candy:New(media:Fetch(mType, db.texture), 200, 14)
	normalAnchor.bars[bar] = true
	bar.candyBarBackground:SetVertexColor(colorBackground())
	bar:Set("hotordot:anchor", normalAnchor)
	local color = DebuffTypeColor[debuffType]
	if color then
		bar:SetColor( color.r, color.g, color.b )
	else
		bar:SetColor( colorNormal() )
	end
	bar.candyBarLabel:SetTextColor(colorText())
	bar.candyBarLabel:SetJustifyH(db.align)
	bar.candyBarLabel:SetFont(media:Fetch("font", db.font), 10)
	bar.candyBarDuration:SetFont(media:Fetch("font", db.font), 10)
	bar:SetLabel(text)
	bar:SetClampedToScreen(true)
	bar:SetDuration(duration)
	bar:SetTimeVisibility(db.time)
	bar:SetIcon(db.icon and icon or nil)
	bar:SetScale(db.scale)
	bar:Start() -- this will set the expiration
	bar:SetDuration(time) -- this will now set the actual time left on the bar
	bar:Set("hotordot:expirationTime", expirationTime)
	bar:Set("hotordot:identifier", text..rank)
	namedBars[text..rank] = bar
end

local done = {}
local function updateBars()
	local dirty = nil
	wipe(done)
	for k, bar in pairs(namedBars) do
		done[k] = true
	end
	local i = 1
	while true do
		local name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable = UnitDebuff("target", i)
		if not name then break end
		rank = rank or ""
		if unitCaster == "player" and duration and duration > 0 and duration < db.minDuration then
			local bar = namedBars[name..rank]
			if bar and bar:Get("hotordot:expirationTime") ~= expirationTime then
				-- kill runnning bar
				bar:Stop()
				dirty = true
				bar = nil
			end
			if not bar then
				-- start bar
				local t = expirationTime - GetTime()
				start(name, rank, icon, t, duration, expirationTime, debuffType)
				done[name..rank] = nil
				dirty = true
			end
			done[name..rank] = nil
		end
		i = i + 1
	end
	for k, v in pairs(done) do
		local bar = namedBars[k]
		if bar then 
			bar:Stop()
			dirty = true
		end
	end
	if dirty then
		rearrangeBars(normalAnchor)
	end
end

function addon:PLAYER_TARGET_CHANGED()
	updateBars()
end

function addon:UNIT_AURA(event, unit)
	if unit == "target" then updateBars() end
end

-- CONFIG

